
Imports System.Collections
Imports System.Diagnostics

' ------------------------------------------------------------------ a "priority queue" data structure
Public Class PriorityQueue
    Implements ICollection

#Region "PriorityQueue Members"

    ' --------------------------------------------------------- maximum level of the skip list
    Private Const LevelMaxValue As Integer = 16
    ' --------------------------------------------------------- probability value used to randomly select the next level value
    Private Const Probability As Double = 0.5
    ' --------------------------------------------------------- current level of the skip list
    Private currentLevel As Integer = 1
    ' --------------------------------------------------------- header node of the skip list
    Private header As New Node(Nothing, LevelMaxValue)
    ' --------------------------------------------------------- Used to generate node levels
    Private rand As New Random()
    ' --------------------------------------------------------- number of elements in the PriorityQueue
    Private m_count As Integer = 0
    ' --------------------------------------------------------- version of this PriorityQueue
    Private version As Long = 0
    ' --------------------------------------------------------- used for comparing and sorting elements
    Private comparer As IComparer


    Public Sub New()
        comparer = New DefaultComparer()
    End Sub

    ' --------------------------------------------------------- new instance with the specified IComparer
    Public Sub New(ByVal comparer As IComparer)
        ' If no comparer was provided.
        If comparer Is Nothing Then
            ' Use the DefaultComparer.
            Me.comparer = New DefaultComparer()
        Else
            ' Else a comparer was provided.
            ' Use the provided comparer.
            Me.comparer = comparer
        End If
    End Sub



#Region "Methods"

    ' ----------------------------------------------------- enqueues the specified element into the PriorityQueue
    Public Overridable Sub Enqueue(ByVal element As Object)
        If element Is Nothing Then
            Throw New ArgumentNullException("element")
        End If

        Dim x As Node = header
        Dim update As Node() = New Node(LevelMaxValue - 1) {}
        Dim nextLevel__1 As Integer = NextLevel()

        ' Find the place in the queue to insert the new element.
        For i As Integer = currentLevel - 1 To 0 Step -1
            While x(i) IsNot Nothing AndAlso comparer.Compare(x(i).Element, element) > 0
                x = x(i)
            End While

            update(i) = x
        Next

        ' If the new node's level is greater than the current level.
        If nextLevel__1 > currentLevel Then
            For i As Integer = currentLevel To nextLevel__1 - 1
                update(i) = header
            Next

            ' Update level.
            currentLevel = nextLevel__1
        End If

        ' Create new node.
        Dim newNode As New Node(element, nextLevel__1)

        ' Insert the new node into the list.
        For i As Integer = 0 To nextLevel__1 - 1
            newNode(i) = update(i)(i)
            update(i)(i) = newNode
        Next

        ' Keep track of the number of elements in the PriorityQueue.
        m_count += 1

        version += 1

        Debug.Assert(Contains(element), "Contains Test", "Contains test for element " & element.ToString() & " failed.")
        AssertValid()
    End Sub


    ' -------------------------------------------------------- removes the element at the head of the PriorityQueue
    Public Overridable Function Dequeue() As Object
        If Count = 0 Then
            Throw New InvalidOperationException("Cannot dequeue into an empty PriorityQueue.")
        End If

        ' Get the first item in the queue.
        Dim element As Object = header(0).Element

        ' Keep track of the node that is about to be removed.
        Dim oldNode As Node = header(0)

        ' Update the header so that its pointers that pointed to the
        ' node to be removed now point to the node that comes after it.
        Dim i As Integer = 0
        While i < currentLevel AndAlso header(i) Is oldNode
            header(i) = oldNode(i)
            i += 1
        End While

        ' Update the current level of the list in case the node that
        ' was removed had the highest level.
        While currentLevel > 1 AndAlso header(currentLevel - 1) Is Nothing
            currentLevel -= 1
        End While

        ' Keep track of how many items are in the queue.
        m_count -= 1

        version += 1

        Debug.Assert(m_count >= 0)
        AssertValid()

        Return element
    End Function


    ' ------------------------------------------------------ removes the specified element from the PriorityQueue
    Public Overridable Sub Remove(ByVal element As Object)
        If element Is Nothing Then
            Throw New ArgumentNullException("element")
        End If

        Dim x As Node = header
        Dim update As Node() = New Node(LevelMaxValue - 1) {}
        Dim nextLevel__1 As Integer = NextLevel()

        ' Find the specified element.
        For i As Integer = currentLevel - 1 To 0 Step -1
            While x(i) IsNot Nothing AndAlso comparer.Compare(x(i).Element, element) > 0
                x = x(i)
            End While

            update(i) = x
        Next

        x = x(0)

        ' If the specified element was found.
        If x IsNot Nothing AndAlso comparer.Compare(x.Element, element) = 0 Then
            ' Remove element.
            Dim i As Integer = 0
            While i < currentLevel AndAlso update(i)(i) Is x
                update(i)(i) = x(i)
                i += 1
            End While

            ' Update list level.
            While currentLevel > 1 AndAlso header(currentLevel - 1) Is Nothing
                currentLevel -= 1
            End While

            ' Keep track of the number of elements in the PriorityQueue.
            m_count -= 1

            version += 1
        End If

        AssertValid()
    End Sub


    ' --------------------------------------------------------- test if the specified element is in the PriorityQueue
    '
    Public Overridable Function Contains(ByVal element As Object) As Boolean
        If element Is Nothing Then
            Return False
        End If

        Dim found As Boolean
        Dim x As Node = header

        ' Find the specified element.
        For i As Integer = currentLevel - 1 To 0 Step -1
            While x(i) IsNot Nothing AndAlso comparer.Compare(x(i).Element, element) > 0
                x = x(i)
            End While
        Next

        x = x(0)

        ' If the element is in the PriorityQueue.
        If x IsNot Nothing AndAlso comparer.Compare(x.Element, element) = 0 Then
            found = True
        Else
            ' Else the element is not in the PriorityQueue.
            found = False
        End If

        Return found
    End Function


    ' ------------------------------------------- returns element at the head of the PriorityQueue without removing it
    '
    Public Overridable Function Peek() As Object
        If Count = 0 Then
            Throw New InvalidOperationException("Cannot peek into an empty PriorityQueue.")
        End If
        Return header(0).Element
    End Function


    ' ------------------------------------------- removes all elements from the PriorityQueue
    Public Overridable Sub Clear()
        header = New Node(Nothing, LevelMaxValue)

        currentLevel = 1

        m_count = 0

        version += 1

        AssertValid()
    End Sub


    ' ------------------------------------------------- returns a synchronized wrapper of the specified PriorityQueue
    '
    Public Shared Function Synchronized(ByVal queue As PriorityQueue) As PriorityQueue
        If queue Is Nothing Then
            Throw New ArgumentNullException("queue")
        End If
        Return New SynchronizedPriorityQueue(queue)
    End Function

    ' ------------------------------------------------- generates a random level for the next node
    Private Function NextLevel() As Integer
        Dim _nextLevel As Integer = 1
        While rand.NextDouble() < Probability AndAlso _nextLevel < LevelMaxValue AndAlso _nextLevel <= currentLevel
            _nextLevel += 1
        End While
        Return _nextLevel
    End Function

    ' ------------------------------------------------- test if PriorityQueue's invariants have been violated
    <Conditional("DEBUG")> _
    Private Sub AssertValid()
        Dim n As Integer = 0
        Dim x As Node = header(0)

        While x IsNot Nothing
            If x(0) IsNot Nothing Then
                Debug.Assert(comparer.Compare(x.Element, x(0).Element) >= 0, "Order test")
            End If

            x = x(0)
            n += 1
        End While

        Debug.Assert(n = Count, "Count test.")

        For i As Integer = 1 To currentLevel - 1
            Debug.Assert(header(i) IsNot Nothing, "Level non-null test.")
        Next

        For i As Integer = currentLevel To LevelMaxValue - 1
            Debug.Assert(header(i) Is Nothing, "Level null test.")
        Next
    End Sub

    <Conditional("DEBUG")> _
    Public Shared Sub Test()
        Dim r As New Random()
        Dim queue As New PriorityQueue()
        Dim count As Integer = 1000
        Dim element As Integer

        For i As Integer = 0 To count - 1
            element = r.[Next]()
            queue.Enqueue(element)

            Debug.Assert(queue.Contains(element), "Contains Test")
        Next

        Debug.Assert(queue.Count = count, "Count Test")

        Dim previousElement As Integer = CInt(queue.Peek())
        Dim peekElement As Integer

        For i As Integer = 0 To count - 1
            peekElement = CInt(queue.Peek())
            element = CInt(queue.Dequeue())

            Debug.Assert(element = peekElement, "Peek Test")
            Debug.Assert(element <= previousElement, "Order Test")

            previousElement = element
        Next

        Debug.Assert(queue.Count = 0)
    End Sub

#End Region


#Region "Private Classes"


#Region "SynchronizedPriorityQueue Class"

    ' -------------------------------------------------------- synchronized wrapper for the PriorityQueue class
    Private Class SynchronizedPriorityQueue
        Inherits PriorityQueue
        Private queue As PriorityQueue

        Private root As Object

        Public Sub New(ByVal queue As PriorityQueue)
            If queue Is Nothing Then
                Throw New ArgumentNullException("queue")
            End If
            Me.queue = queue
            root = queue.SyncRoot
        End Sub

        Public Overloads Overrides Sub Enqueue(ByVal element As Object)
            SyncLock root
                queue.Enqueue(element)
            End SyncLock
        End Sub

        Public Overloads Overrides Function Dequeue() As Object
            SyncLock root
                Return queue.Dequeue()
            End SyncLock
        End Function

        Public Overloads Overrides Sub Remove(ByVal element As Object)
            SyncLock root
                queue.Remove(element)
            End SyncLock
        End Sub

        Public Overloads Overrides Sub Clear()
            SyncLock root
                queue.Clear()
            End SyncLock
        End Sub

        Public Overloads Overrides Function Contains(ByVal element As Object) As Boolean
            SyncLock root
                Return queue.Contains(element)
            End SyncLock
        End Function

        Public Overloads Overrides Function Peek() As Object
            SyncLock root
                Return queue.Peek()
            End SyncLock
        End Function

        Public Overloads Overrides Sub CopyTo(ByVal array As Array, ByVal index As Integer)
            SyncLock root
                queue.CopyTo(array, index)
            End SyncLock
        End Sub

        Public Overloads Overrides ReadOnly Property Count() As Integer
            Get
                SyncLock root
                    Return queue.Count
                End SyncLock
            End Get
        End Property

        Public Overloads Overrides ReadOnly Property IsSynchronized() As Boolean
            Get
                Return True
            End Get
        End Property

        Public Overloads Overrides ReadOnly Property SyncRoot() As Object
            Get
                Return root
            End Get
        End Property

        Public Overloads Overrides Function GetEnumerator() As IEnumerator
            SyncLock root
                Return queue.GetEnumerator()
            End SyncLock
        End Function
    End Class

#End Region

#Region "DefaultComparer Class"

    ' ------------------------------------------------- IComparer to use if no comparer was provided
    Private Class DefaultComparer
        Implements IComparer

        Public Function Compare(ByVal x As Object, ByVal y As Object) As Integer Implements IComparer.Compare
            If Not (TypeOf y Is IComparable) Then
                Throw New ArgumentException("Item does not implement IComparable.")
            End If

            Dim a As IComparable = TryCast(x, IComparable)

            Debug.Assert(a IsNot Nothing)

            Return a.CompareTo(y)
        End Function

    End Class

#End Region

#Region "Node Class"

    ' ---------------------------------------------------------- a node in the list of nodes.
    Private Class Node
        Private forward As Node()

        Private m_element As Object

        Public Sub New(ByVal element As Object, ByVal level As Integer)
            Me.forward = New Node(level - 1) {}
            Me.m_element = element
        End Sub

        Default Public Property Item(ByVal index As Integer) As Node
            Get
                Return forward(index)
            End Get
            Set(ByVal value As Node)
                forward(index) = value
            End Set
        End Property

        Public ReadOnly Property Element() As Object
            Get
                Return m_element
            End Get
        End Property
    End Class

#End Region

#Region "PriorityQueueEnumerator Class"

    ' ---------------------------------------------- implements IEnumerator interface for the PriorityQueue class
    Private Class PriorityQueueEnumerator
        Implements IEnumerator
        Private owner As PriorityQueue

        Private head As Node

        Private currentNode As Node

        Private moveResult As Boolean

        Private version As Long

        Public Sub New(ByVal owner As PriorityQueue)
            Me.owner = owner
            Me.version = owner.version
            head = owner.header

            Reset()
        End Sub

        Public Sub Reset() Implements IEnumerator.Reset
            If version <> owner.version Then
                Throw New InvalidOperationException("The PriorityQueue was modified after the enumerator was created.")
            End If

            currentNode = head
            moveResult = True
        End Sub

        Public ReadOnly Property Current() As Object Implements IEnumerator.Current
            Get
                If currentNode Is head OrElse currentNode Is Nothing Then
                    Throw New InvalidOperationException("The enumerator is positioned before the first " & "element of the collection or after the last element.")
                End If

                Return currentNode.Element
            End Get
        End Property

        Public Function MoveNext() As Boolean Implements IEnumerator.MoveNext
            If version <> owner.version Then
                Throw New InvalidOperationException("The PriorityQueue was modified after the enumerator was created.")
            End If

            If moveResult Then
                currentNode = currentNode(0)
            End If

            If currentNode Is Nothing Then
                moveResult = False
            End If

            Return moveResult
        End Function

    End Class

#End Region

#End Region

#End Region


#Region "ICollection Members"

    Public Overridable ReadOnly Property IsSynchronized() As Boolean Implements ICollection.IsSynchronized
        Get
            Return False
        End Get
    End Property

    Public Overridable ReadOnly Property Count() As Integer Implements ICollection.Count
        Get
            Return m_count
        End Get
    End Property

    Public Overridable Sub CopyTo(ByVal array As Array, ByVal index As Integer) Implements ICollection.CopyTo
        If array Is Nothing Then
            Throw New ArgumentNullException("array")
        ElseIf index < 0 Then
            Throw New ArgumentOutOfRangeException("index", index, "Array index out of range.")
        ElseIf array.Rank > 1 Then
            Throw New ArgumentException("Array has more than one dimension.", "array")
        ElseIf index >= array.Length Then
            Throw New ArgumentException("index is equal to or greater than the length of array.", "index")
        ElseIf Count > array.Length - index Then
            Throw New ArgumentException("The number of elements in the PriorityQueue is greater " & "than the available space from index to the end of the " & "destination array.", "index")
        End If

        Dim i As Integer = index

        For Each element As Object In Me
            array.SetValue(element, i)
            i += 1
        Next
    End Sub

    Public Overridable ReadOnly Property SyncRoot() As Object Implements ICollection.SyncRoot
        Get
            Return Me
        End Get
    End Property

#End Region

#Region "IEnumerable Members"

    Public Overridable Function GetEnumerator() As IEnumerator Implements IEnumerable.GetEnumerator
        Return New PriorityQueueEnumerator(Me)
    End Function

#End Region

End Class
