

Imports System.IO

' -------------------------------------------------------------------------- reads a track from a stream
Friend Class TrackReader

    Private m_track As New Track()
    Private newTrack As New Track()
    Private cmBuilder As New ChannelMessageBuilder()
    Private scBuilder As New SysCommonMessageBuilder()
    Private stream As Stream
    Private trackData As Byte()
    Private trackIndex As Integer
    Private previousTicks As Integer
    Private ticks As Integer
    Private status As Integer
    Private runningStatus As Integer

    Public Sub New()
    End Sub

    Public Sub Read(ByVal strm As Stream)
        stream = strm
        FindTrack()

        Dim trackLength As Integer = GetTrackLength()
        trackData = New Byte(trackLength - 1) {}

        Dim result As Integer = strm.Read(trackData, 0, trackLength)

        If result < 0 Then
            Throw New MidiFileException("End of MIDI file unexpectedly reached.")
        End If

        newTrack = New Track()
        ParseTrackData()
        m_track = newTrack
    End Sub

    Private Sub FindTrack()
        Dim found As Boolean = False
        Dim result As Integer

        While Not found
            result = stream.ReadByte()

            If result = Asc("M"c) Then
                result = stream.ReadByte()

                If result = Asc("T"c) Then
                    result = stream.ReadByte()

                    If result = Asc("r"c) Then
                        result = stream.ReadByte()

                        If result = Asc("k"c) Then
                            found = True
                        End If
                    End If
                End If
            End If

            If result < 0 Then
                Throw New MidiFileException("Unable to find track in MIDI file.")
            End If
        End While
    End Sub

    Private Function GetTrackLength() As Integer
        Dim trackLength As Byte() = New Byte(3) {}

        Dim result As Integer = stream.Read(trackLength, 0, trackLength.Length)

        If result < trackLength.Length Then
            Throw New MidiFileException("End of MIDI file unexpectedly reached.")
        End If

        If BitConverter.IsLittleEndian Then
            Array.Reverse(trackLength)
        End If

        Return BitConverter.ToInt32(trackLength, 0)
    End Function

    Private Sub ParseTrackData()
        trackIndex = 0
        ticks = 0
        runningStatus = 0

        While trackIndex < trackData.Length
            previousTicks = ticks

            ticks += ReadVariableLengthValue()

            If (trackData(trackIndex) And &H80) = &H80 Then
                status = trackData(trackIndex)
                trackIndex += 1
            Else
                status = runningStatus
            End If

            ParseMessage()
        End While
    End Sub

    Private Sub ParseMessage()
        ' ---------------------------------------------------------------- channel message
        If status >= ChannelCommand.NoteOff AndAlso _
            status <= ChannelCommand.PitchWheel + ChannelMessage.MidiChannelMaxValue Then
            ParseChannelMessage()
            ' ------------------------------------------------------------ meta message
        ElseIf status = &HFF Then
            ParseMetaMessage()
            ' ------------------------------------------------------------ start of a system exclusive message
        ElseIf status = SysExType.Start Then
            ParseSysExMessageStart()
            ' ------------------------------------------------------------ continuation of a system exclusive message
        ElseIf status = SysExType.Continuation Then
            ParseSysExMessageContinue()
            ' ------------------------------------------------------------ system common message
        ElseIf status >= SysCommonType.MidiTimeCode AndAlso _
                status <= SysCommonType.TuneRequest Then
            ParseSysCommonMessage()
            ' ------------------------------------------------------------ system realtime message
        ElseIf status >= SysRealtimeType.MsgClock AndAlso _
                status <= SysRealtimeType.MsgReset Then
            ParseSysRealtimeMessage()
        End If
    End Sub

    Private Sub ParseChannelMessage()
        If trackIndex >= trackData.Length Then
            Throw New MidiFileException("End of track unexpectedly reached.")
        End If

        cmBuilder.Command = ChannelMessage.UnpackCommand(status)
        cmBuilder.MidiChannel = ChannelMessage.UnpackMidiChannel(status)
        cmBuilder.Data1 = trackData(trackIndex)

        trackIndex += 1

        If ChannelMessage.DataBytesPerType(cmBuilder.Command) = 2 Then
            If trackIndex >= trackData.Length Then
                Throw New MidiFileException("End of track unexpectedly reached.")
            End If

            cmBuilder.Data2 = trackData(trackIndex)

            trackIndex += 1
        End If

        cmBuilder.Build()

        newTrack.Insert(ticks, cmBuilder.Result)
        runningStatus = status
    End Sub

    Private Sub ParseMetaMessage()
        If trackIndex >= trackData.Length Then
            Throw New MidiFileException("End of track unexpectedly reached.")
        End If

        Dim type As MetaType = CType(trackData(trackIndex), MetaType)

        trackIndex += 1

        If trackIndex >= trackData.Length Then
            Throw New MidiFileException("End of track unexpectedly reached.")
        End If

        If type = MetaType.EndOfTrack Then
            newTrack.EndOfTrackOffset = ticks - previousTicks

            trackIndex += 1
        Else
            Dim data As Byte() = New Byte(ReadVariableLengthValue() - 1) {}
            Array.Copy(trackData, trackIndex, data, 0, data.Length)
            newTrack.Insert(ticks, New MetaMessage(type, data))

            trackIndex += data.Length
        End If
    End Sub

    Private Sub ParseSysExMessageStart()
        ' ---------------------------------------------------------------- system exclusive cancels running status
        runningStatus = 0

        Dim data As Byte() = New Byte(ReadVariableLengthValue()) {}
        data(0) = CByte(SysExType.Start)

        Array.Copy(trackData, trackIndex, data, 1, data.Length - 1)
        newTrack.Insert(ticks, New SysExMessage(data))

        trackIndex += data.Length - 1
    End Sub

    Private Sub ParseSysExMessageContinue()
        trackIndex += 1

        If trackIndex >= trackData.Length Then
            Throw New MidiFileException("End of track unexpectedly reached.")
        End If
        ' ---------------------------------------------------------------- system exclusive cancels running status
        runningStatus = 0
        ' ---------------------------------------------------------------- test if this is an escaped message rather 
        '                                                                  than a system exclusive continuation message.
        If (trackData(trackIndex) And &H80) = &H80 Then
            status = trackData(trackIndex)
            trackIndex += 1

            ParseMessage()
        Else
            Dim data As Byte() = New Byte(ReadVariableLengthValue()) {}
            data(0) = CByte(SysExType.Continuation)

            Array.Copy(trackData, trackIndex, data, 1, data.Length - 1)
            newTrack.Insert(ticks, New SysExMessage(data))

            trackIndex += data.Length - 1
        End If
    End Sub

    Private Sub ParseSysCommonMessage()
        If trackIndex >= trackData.Length Then
            Throw New MidiFileException("End of track unexpectedly reached.")
        End If
        ' ---------------------------------------------------------------- system exclusive cancels running status
        runningStatus = 0

        scBuilder.Type = CType(status, SysCommonType)

        Select Case CType(status, SysCommonType)
            Case SysCommonType.MidiTimeCode
                scBuilder.Data1 = trackData(trackIndex)
                trackIndex += 1

            Case SysCommonType.SongPositionPointer
                scBuilder.Data1 = trackData(trackIndex)
                trackIndex += 1

                If trackIndex >= trackData.Length Then
                    Throw New MidiFileException("End of track unexpectedly reached.")
                End If

                scBuilder.Data2 = trackData(trackIndex)
                trackIndex += 1

            Case SysCommonType.SongSelect
                scBuilder.Data1 = trackData(trackIndex)
                trackIndex += 1

            Case SysCommonType.TuneRequest

        End Select

        scBuilder.Build()

        newTrack.Insert(ticks, scBuilder.Result)
    End Sub

    Private Sub ParseSysRealtimeMessage()
        Dim e As SysRealtimeMessage = Nothing

        Select Case CType(status, SysRealtimeType)
            Case SysRealtimeType.MsgActiveSense : e = SysRealtimeMessage.ActiveSenseMessage

            Case SysRealtimeType.MsgClock : e = SysRealtimeMessage.ClockMessage

            Case SysRealtimeType.MsgContinue : e = SysRealtimeMessage.ContinueMessage

            Case SysRealtimeType.MsgReset : e = SysRealtimeMessage.ResetMessage

            Case SysRealtimeType.MsgStart : e = SysRealtimeMessage.StartMessage

            Case SysRealtimeType.MsgStop : e = SysRealtimeMessage.StopMessage

            Case SysRealtimeType.MsgTick : e = SysRealtimeMessage.TickMessage
        End Select

        newTrack.Insert(ticks, e)
    End Sub

    Private Function ReadVariableLengthValue() As Integer
        If trackIndex >= trackData.Length Then
            Throw New MidiFileException("End of track unexpectedly reached.")
        End If

        Dim result As Integer = 0

        result = trackData(trackIndex)

        trackIndex += 1

        If (result And &H80) = &H80 Then
            result = result And &H7F

            Dim temp As Integer

            Do
                If trackIndex >= trackData.Length Then
                    Throw New MidiFileException("End of track unexpectedly reached.")
                End If

                temp = trackData(trackIndex)
                trackIndex += 1
                result <<= 7
                result = result Or temp And &H7F
            Loop While (temp And &H80) = &H80
        End If

        Return result
    End Function

    Public ReadOnly Property Track() As Track
        Get
            Return m_track
        End Get
    End Property

End Class
