

Imports System.Diagnostics
Imports LC.Midi

' -------------------------------------------------------------------- low frequency oscillator
Public Class Lfo
	Inherits MonoSynthComponent
	Implements IProgrammable
	Implements IControllable

    Public Enum ParameterId
        Shape
        Speed
        Delay
        ModulationWheelEnabled
    End Enum

    Public Enum Shape
        Sine
        Triangle
        Square
        DownRamp
        UpRamp
        SampleAndHold
    End Enum


    Private Const SpeedScaler As Integer = 20
    Private Const ShapeCount As Integer = Shape.SampleAndHold + 1
    Private Const DelayScaler As Integer = 3

    Private rand As New Random()
    Private shOutput As Single = 0
    Private speed As Single
    Private m_shape As Shape
    Private delay As Integer
    Private sine As Single = 0
    Private cosine As Single = 1
    Private amplitude As Single = 1
    Private modulationWheelEnabled As Boolean = False
    Private samples As Integer = 0
    Private increment As Single
    Private accumulator As Single = 0
    Private m_synthesizeReplaceEnabled As Boolean = True



    Public Sub New(ByVal sampleRate As SampleRate, ByVal buffer As MonoBuffer)
        MyBase.New(sampleRate, buffer)
        If sampleRate Is Nothing Then
            Throw New ArgumentNullException("sampleRate")
        ElseIf buffer Is Nothing Then
            Throw New ArgumentNullException("buffer")
        End If
        AddHandler sampleRate.SampleRateChanged, AddressOf HandleSampleRateChanged
        InitializeParameters()
    End Sub

    Public Sub New(ByVal sampleRate As SampleRate, ByVal buffer As MonoBuffer, ByVal name As String)
        MyBase.New(sampleRate, buffer, name)
        '
        If sampleRate Is Nothing Then
            Throw New ArgumentNullException("sampleRate")
        ElseIf buffer Is Nothing Then
            Throw New ArgumentNullException("buffer")
        ElseIf name Is Nothing Then
            Throw New ArgumentNullException("name")
        End If
        AddHandler sampleRate.SampleRateChanged, AddressOf HandleSampleRateChanged
        InitializeParameters()
    End Sub


    Private Sub InitializeParameters()
        SetParameterValue(ParameterId.Shape, 0)
        SetParameterValue(ParameterId.Delay, 0)
        SetParameterValue(ParameterId.Speed, 0.2F)
    End Sub

    Private Sub CalculatePhaseIncrement()
        increment = PI2 * speed / SamplesPerSecond
    End Sub

    Public Overloads Overrides Sub Synthesize(ByVal offset As Integer, ByVal count As Integer)
        Dim buffer As Single() = GetBuffer()

        Debug.Assert(buffer.Length >= offset + count)

        Dim start As Integer = offset
        Dim [end] As Integer = offset + count

        ' ------------------------------------------------ While the end of the delay has not been reached and
        '                                                  the start point hasn't reached the end point
        While samples < delay AndAlso start < [end]
            If m_synthesizeReplaceEnabled Then
                buffer(start) = 0
            End If

            samples += 1
            start += 1
        End While

        If samples >= delay Then
            Select Case m_shape
                Case Shape.Sine : SynthesizeSine(start, [end])
                Case Shape.Triangle : SynthesizeTriangle(start, [end])
                Case Shape.Square : SynthesizeSquare(start, [end])
                Case Shape.DownRamp : SynthesizeDownRamp(start, [end])
                Case Shape.UpRamp : SynthesizeUpRamp(start, [end])
                Case Shape.SampleAndHold : SynthesizeSampleAndHold(start, [end])
                Case Else : Debug.Fail("Unhandled shape.")
            End Select
        End If

        ' -------------------------------------------------- increase sample counter by the number of remaining samples 
        samples += [end] - start
    End Sub

    Public Overloads Overrides Sub Trigger(ByVal previousNote As Integer, ByVal note As Integer, ByVal velocity As Single)
        samples = 0
        sine = 0
        cosine = 1
        Select Case m_shape
            Case Shape.Sine, Shape.Square, Shape.SampleAndHold : accumulator = 0
            Case Shape.Triangle : accumulator = 0.25F * PI2
            Case Shape.DownRamp, Shape.UpRamp : accumulator = CSng(Math.PI)
            Case Else : Debug.Fail("Unhandled shape.")
        End Select
    End Sub

    Public Overloads Overrides Sub Release(ByVal velocity As Single)
    End Sub

    Private Function IncrementPhase() As Boolean
        Dim result As Boolean
        sine = sine + increment * cosine
        cosine = cosine - increment * sine
        accumulator += increment
        If accumulator >= PI2 Then
            accumulator -= PI2
            result = True
        Else
            result = False
        End If
        Return result
    End Function

    Private Sub SynthesizeSine(ByVal start As Integer, ByVal [end] As Integer)
        Dim buffer As Single() = GetBuffer()
        Dim output As Single
        For i As Integer = start To [end] - 1
            output = sine * amplitude
            buffer(i) = (If(m_synthesizeReplaceEnabled, output, buffer(i) + output))
            IncrementPhase()
        Next
    End Sub

    Private Sub SynthesizeTriangle(ByVal start As Integer, ByVal [end] As Integer)
        Dim buffer As Single() = GetBuffer()
        Dim output As Single
        For i As Integer = start To [end] - 1
            If accumulator < Math.PI Then
                output = 1 - 4 * accumulator / PI2
            Else
                output = 1 - 4 * (1 - accumulator / PI2)
            End If
            output *= amplitude
            buffer(i) = (If(m_synthesizeReplaceEnabled, output, buffer(i) + output))
            IncrementPhase()
        Next
    End Sub

    Private Sub SynthesizeSquare(ByVal start As Integer, ByVal [end] As Integer)
        Dim buffer As Single() = GetBuffer()
        Dim output As Single
        For i As Integer = start To [end] - 1
            If accumulator < Math.PI Then
                output = 1
            Else
                output = -1
            End If
            output *= amplitude
            buffer(i) = (If(m_synthesizeReplaceEnabled, output, buffer(i) + output))
            IncrementPhase()
        Next
    End Sub

    Private Sub SynthesizeDownRamp(ByVal start As Integer, ByVal [end] As Integer)
        Dim buffer As Single() = GetBuffer()
        Dim output As Single
        For i As Integer = start To [end] - 1
            output = 1 - 2 * accumulator / PI2
            output *= amplitude
            buffer(i) = (If(m_synthesizeReplaceEnabled, output, buffer(i) + output))
            IncrementPhase()
        Next
    End Sub

    Private Sub SynthesizeUpRamp(ByVal start As Integer, ByVal [end] As Integer)
        Dim buffer As Single() = GetBuffer()
        Dim output As Single
        For i As Integer = start To [end] - 1
            output = 1 - 2 * (1 - accumulator / PI2)
            output *= amplitude
            buffer(i) = (If(m_synthesizeReplaceEnabled, output, buffer(i) + output))
            IncrementPhase()
        Next
    End Sub

    Private Sub SynthesizeSampleAndHold(ByVal start As Integer, ByVal [end] As Integer)
        Dim buffer As Single() = GetBuffer()
        Dim output As Single
        For i As Integer = start To [end] - 1
            output = shOutput
            output *= amplitude
            buffer(i) = (If(m_synthesizeReplaceEnabled, output, buffer(i) + output))
            If IncrementPhase() Then
                shOutput = 1 - 2 * CSng(rand.NextDouble())
            End If
        Next
    End Sub

    Private Sub HandleSampleRateChanged(ByVal sender As Object, ByVal e As EventArgs)
        CalculatePhaseIncrement()
    End Sub


    Public Overloads Overrides ReadOnly Property Ordinal() As Integer
        Get
            Return 1
        End Get
    End Property

    Public Overloads Overrides Property SynthesizeReplaceEnabled() As Boolean
        Get
            Return m_synthesizeReplaceEnabled
        End Get
        Set(ByVal value As Boolean)
            m_synthesizeReplaceEnabled = value
        End Set
    End Property



#Region "IProgrammable Members"

    Public Function GetParameterName(ByVal index As Integer) As String Implements IProgrammable.GetParameterName
        If index < 0 OrElse index >= ParameterCount Then
            Throw New ArgumentOutOfRangeException("index")
        End If
        '
        Dim result As String = ""
        Dim name__1 As String = Name
        If Not String.IsNullOrEmpty(name__1) Then name__1 = name__1 & " "
        '
        Select Case CType(index, ParameterId)
            Case ParameterId.Delay : result = name__1 & "Delay"
            Case ParameterId.Shape : result = name__1 & "Shape"
            Case ParameterId.Speed : result = name__1 & "Speed"
            Case ParameterId.ModulationWheelEnabled : result = name__1 & "Modulation Wheel Enabled"
            Case Else : Debug.Fail("Unhandled parameter")
        End Select
        '
        Return result
    End Function

    Public Function GetParameterLabel(ByVal index As Integer) As String Implements IProgrammable.GetParameterLabel
        If index < 0 OrElse index >= ParameterCount Then
            Throw New ArgumentOutOfRangeException("index")
        End If
        Dim result As String = ""
        Select Case CType(index, ParameterId)
            Case ParameterId.Delay : result = "Seconds"
            Case ParameterId.Shape : result = "Shape"
            Case ParameterId.Speed : result = "Hz"
            Case ParameterId.ModulationWheelEnabled : result = "On/Off"
            Case Else : Debug.Fail("Unhandled parameter")
        End Select
        Return result
    End Function

    Public Function GetParameterDisplay(ByVal index As Integer) As String Implements IProgrammable.GetParameterDisplay
        If index < 0 OrElse index >= ParameterCount Then
            Throw New ArgumentOutOfRangeException("index")
        End If
        '
        Dim result As String = ""
        '
        Select Case CType(index, ParameterId)
            Case ParameterId.Delay
                If True Then
                    Dim seconds As Single = CSng(delay) / SamplesPerSecond

                    result = seconds.ToString(Synthesizer.FormatInfo)
                End If

            Case ParameterId.Shape
                result = m_shape.ToString()

            Case ParameterId.Speed
                result = speed.ToString(Synthesizer.FormatInfo)

            Case ParameterId.ModulationWheelEnabled
                result = (If(modulationWheelEnabled, "On", "Off"))

            Case Else
                Debug.Fail("Unhandled parameter")

        End Select
        '
        Return result
    End Function

    Public Function GetParameterValue(ByVal index As Integer) As Single Implements IProgrammable.GetParameterValue
        If index < 0 OrElse index >= ParameterCount Then
            Throw New ArgumentOutOfRangeException("index")
        End If
        '
        Dim result As Single = 0
        '
        Select Case CType(index, ParameterId)
            Case ParameterId.Delay
                result = CSng(delay / DelayScaler / SamplesPerSecond)

            Case ParameterId.Shape
                result = CSng(m_shape / (ShapeCount - 1))

            Case ParameterId.Speed
                result = speed / SpeedScaler

            Case ParameterId.ModulationWheelEnabled
                result = (If(modulationWheelEnabled, 1, 0))

            Case Else
                Debug.Fail("Unhandled parameter")

        End Select
        '
        Return result
    End Function

    Public Sub SetParameterValue(ByVal index As Integer, ByVal value As Single) Implements IProgrammable.SetParameterValue
        If index < 0 OrElse index >= ParameterCount Then
            Throw New ArgumentOutOfRangeException("index")
        End If

        Select Case CType(index, ParameterId)
            Case ParameterId.Delay
                delay = CInt(Math.Truncate(value * DelayScaler * SamplesPerSecond))

            Case ParameterId.Shape
                m_shape = CType(Math.Round(value * (ShapeCount - 1)), Shape)

            Case ParameterId.Speed
                speed = value * SpeedScaler
                CalculatePhaseIncrement()

            Case ParameterId.ModulationWheelEnabled
                modulationWheelEnabled = (If(value < 0.5, False, True))
                If modulationWheelEnabled Then
                    amplitude = 0
                Else
                    amplitude = 1
                End If

            Case Else
                Debug.Fail("Unhandled parameter")

        End Select
    End Sub

    Public ReadOnly Property ParameterCount() As Integer Implements IProgrammable.ParameterCount
        Get
            Return ParameterId.ModulationWheelEnabled + 1
        End Get
    End Property

#End Region

#Region "IControllable Members"

    Public Sub ProcessControllerMessage(ByVal type As ControllerType, ByVal value As Single) Implements IControllable.ProcessControllerMessage
        If type = ControllerType.ModulationWheel Then
            If modulationWheelEnabled Then
                amplitude = value
            End If
        End If
    End Sub

#End Region

End Class
