

Imports System.Diagnostics

Public Class Oscillator
	Inherits MonoSynthComponent
	Implements IProgrammable
    Implements IBendable

	Public Enum ParameterId
		Detune
		Transpose
		AmplitudeLevel
		FMLevel
		FMPolarity
		PortamentoEnabled
		Polarity
	End Enum

	Private Const DetuneScaler As Integer = 100
	Private Const TransposeScaler As Integer = 24
	Private waveforms As Wavetable
	Private waveMap As WaveformMap
	Private currentWaveform As Waveform
    Private currentNote As Integer = 69     ' A-440Hz
	Private accumulator As Single
	Private m_fmModulator As MonoSynthComponent
	Private fmLevel As Single = 0
	Private fmPolarity As Single = 1
	Private portamentoSource As SlewLimiter
	Private portamentoEnabled As Boolean = False
	Private detune As Single = 0
	Private transpose As Integer = 0
	Private m_pitchBendModulation As Single = 0
	Private amplitudeLevel As Single = 1
	Private polarity As Single = 1
	Private m_synthesizeReplaceEnabled As Boolean = True

	Public Sub New(sampleRate As SampleRate, buffer As MonoBuffer, waves As Wavetable, fmModulator As MonoSynthComponent, portamentoSource As SlewLimiter)
		MyBase.New(sampleRate, buffer)

		If waves Is Nothing Then
			Throw New ArgumentNullException("waves")
		ElseIf waves.Count = 0 Then
			Throw New ArgumentException("Wavetable does not contain any waveforms.")
		ElseIf fmModulator Is Nothing Then
			Throw New ArgumentNullException("fmModulator")
		ElseIf portamentoSource Is Nothing Then
			Throw New ArgumentNullException("portamentoSource")
		End If

		Wavetable = waves
		Me.m_fmModulator = fmModulator
		Me.portamentoSource = portamentoSource
	End Sub

	Public Sub New(sampleRate As SampleRate, buffer As MonoBuffer, name As String, waves As Wavetable, fmModulator As MonoSynthComponent, portamentoSource As SlewLimiter)
		MyBase.New(sampleRate, buffer, name)

		If waves Is Nothing Then
			Throw New ArgumentNullException("waves")
		ElseIf waves.Count = 0 Then
			Throw New ArgumentException("Wavetable does not contain any waveforms.")
		ElseIf fmModulator Is Nothing Then
			Throw New ArgumentNullException("fmModulator")
		ElseIf portamentoSource Is Nothing Then
			Throw New ArgumentNullException("portamentoSource")
		End If

		Wavetable = waves
		Me.m_fmModulator = fmModulator
		Me.portamentoSource = portamentoSource
	End Sub

	Public Overloads Overrides Sub Synthesize(offset As Integer, count As Integer)
		Dim buffer As Single() = GetBuffer()
		Dim c As Integer = offset + count
        Dim bendNote As Single = currentNote + 48 * m_pitchBendModulation + detune + transpose
		Dim modNote As Single
		Dim fmBuffer As Single() = m_fmModulator.GetBuffer()
		Dim portamentoBuffer As Single() = portamentoSource.GetBuffer()
		Dim o As Single
		Dim sampleRate As Integer = SamplesPerSecond
        Dim freq As Single

		Debug.Assert(fmBuffer.Length = buffer.Length)

		For i As Integer = offset To c - 1
            modNote = 12 * fmBuffer(i) * (fmLevel * fmLevel) + bendNote

			If portamentoEnabled Then
				modNote += 12 * portamentoBuffer(i)
			End If

			If modNote < 0 Then
				modNote = 0
			ElseIf modNote >= NoteCount Then
				modNote = NoteCount - 1
			End If

			freq = PowerOfTwoTable.GetPower((modNote - A440NoteNumber) / NotesPerOctave) * A440Frequency

			accumulator += freq * currentWaveform.Length / sampleRate

			While accumulator >= currentWaveform.Length
				accumulator -= currentWaveform.Length
			End While

            o = currentWaveform(CInt(Math.Truncate(accumulator))) * amplitudeLevel * polarity

			buffer(i) = (If(m_synthesizeReplaceEnabled, o, buffer(i) + o))
		Next
	End Sub

	Public Overloads Overrides Sub Trigger(previousNote As Integer, note As Integer, velocity As Single)
		currentNote = note
		currentWaveform = waveMap(currentNote + transpose)
	End Sub

	Public Overloads Overrides Sub Release(velocity As Single)
	End Sub

	Public Property FMModulator() As MonoSynthComponent
		Get
			Return m_fmModulator
		End Get
		Set
            If Value Is Nothing Then
                Throw New ArgumentNullException("value")
            End If

            m_fmModulator = Value
            Dim oldOrdinal As Integer = Ordinal
            If oldOrdinal <> Ordinal Then
                OnOrdinalValueChanged(EventArgs.Empty)
            End If
        End Set
	End Property

	Public Property Wavetable() As Wavetable
		Get
			Return waveforms
		End Get
		Set
            If Value Is Nothing Then
                Throw New ArgumentNullException("value")
            End If

            waveforms = Value
            waveMap = New WaveformMap(Value)
            currentWaveform = waveMap(currentNote + transpose)
        End Set
	End Property

	Public Overloads Overrides ReadOnly Property Ordinal() As Integer
		Get
			Return 1 + m_fmModulator.Ordinal + portamentoSource.Ordinal
		End Get
	End Property

	Public Overloads Overrides Property SynthesizeReplaceEnabled() As Boolean
		Get
			Return m_synthesizeReplaceEnabled
		End Get
		Set
			m_synthesizeReplaceEnabled = value
		End Set
	End Property


#Region "IProgrammable Members"

    Public Function GetParameterName(ByVal index As Integer) As String Implements IProgrammable.GetParameterName
        If index < 0 OrElse index >= ParameterCount Then
            Throw New ArgumentOutOfRangeException("index")
        End If

        Dim result As String = ""
        Dim name__1 As String = Name
        If Not String.IsNullOrEmpty(name__1) Then name__1 = name__1 & " "

        Select Case CType(index, ParameterId)
            Case ParameterId.AmplitudeLevel : result = name__1 & "Amplitude Level"
            Case ParameterId.Detune : result = name__1 & "Detune"
            Case ParameterId.Transpose : result = name__1 & "Transpose"
            Case ParameterId.FMLevel : result = name__1 & "FM Level"
            Case ParameterId.FMPolarity : result = name__1 & "FM Polarity"
            Case ParameterId.PortamentoEnabled : result = name__1 & "Portamento Enabled"
            Case ParameterId.Polarity : result = name__1 & "Polarity"
            Case Else : Debug.Fail("Unhandled parameter.")
        End Select

        Return result
    End Function

    Public Function GetParameterLabel(ByVal index As Integer) As String Implements IProgrammable.GetParameterLabel
        If index < 0 OrElse index >= ParameterCount Then
            Throw New ArgumentOutOfRangeException("index")
        End If
        Dim result As String = ""
        Select Case CType(index, ParameterId)
            Case ParameterId.AmplitudeLevel : result = "dB"
            Case ParameterId.Detune : result = "Cents"
            Case ParameterId.Transpose : result = "Semitones"
            Case ParameterId.FMLevel : result = "dB"
            Case ParameterId.FMPolarity : result = "+/-"
            Case ParameterId.PortamentoEnabled : result = "On/Off"
            Case ParameterId.Polarity : result = "+/-"
            Case Else : Debug.Fail("Unhandled parameter.")
        End Select
        Return result
    End Function

    Public Function GetParameterDisplay(ByVal index As Integer) As String Implements IProgrammable.GetParameterDisplay
        If index < 0 OrElse index >= ParameterCount Then
            Throw New ArgumentOutOfRangeException("index")
        End If

        Dim result As String = ""

        Select Case CType(index, ParameterId)
            Case ParameterId.AmplitudeLevel
                result = Synthesizer.DecibelToString(amplitudeLevel)

            Case ParameterId.Detune
                If True Then
                    Dim d As Single = detune * DetuneScaler

                    result = d.ToString(Synthesizer.FormatInfo)
                End If

            Case ParameterId.Transpose
                result = transpose.ToString(Synthesizer.FormatInfo)
   
            Case ParameterId.FMLevel
                result = Synthesizer.DecibelToString(fmLevel)
 
            Case ParameterId.FMPolarity
                result = (If(fmPolarity < 0, "-", "+"))
                
            Case ParameterId.PortamentoEnabled
                result = (If(portamentoEnabled, "On", "Off"))
               
            Case ParameterId.Polarity
                If polarity >= 0 Then
                    result = "+"
                Else
                    result = "-"
                End If

            Case Else
                Debug.Fail("Unhandled parameter.")

        End Select

        Return result
    End Function

    Public Function GetParameterValue(ByVal index As Integer) As Single Implements IProgrammable.GetParameterValue
        If index < 0 OrElse index >= ParameterCount Then
            Throw New ArgumentOutOfRangeException("index")
        End If

        Dim result As Single = 0

        Select Case CType(index, ParameterId)
            Case ParameterId.AmplitudeLevel
                result = amplitudeLevel

            Case ParameterId.Detune
                result = (detune + 1) / 2
                
            Case ParameterId.Transpose
                result = CSng((transpose + TransposeScaler)) / (TransposeScaler * 2)
                
            Case ParameterId.FMLevel
                result = Math.Abs(fmLevel)
               
            Case ParameterId.FMPolarity
                result = (If(fmPolarity < 0, 1, 0))
                
            Case ParameterId.PortamentoEnabled
                result = (If(portamentoEnabled, 1, 0))
               
            Case ParameterId.Polarity
                result = (If(polarity >= 0, 0, 1))

            Case Else
                Debug.Fail("Unhandled parameter.")

        End Select

        Return result
    End Function

    Public Sub SetParameterValue(ByVal index As Integer, ByVal value As Single) Implements IProgrammable.SetParameterValue
     
        If index < 0 OrElse index >= ParameterCount Then
            Throw New ArgumentOutOfRangeException("index")
        End If

        Select Case CType(index, ParameterId)
            Case ParameterId.AmplitudeLevel
                amplitudeLevel = value

            Case ParameterId.Detune
                detune = value * 2 - 1
                
            Case ParameterId.Transpose
                transpose = CInt(Math.Round(value * TransposeScaler * 2 - TransposeScaler))
                
            Case ParameterId.FMLevel
                fmLevel = value * fmPolarity
                
            Case ParameterId.FMPolarity
                fmPolarity = (If(value < 0.5, 1, -1))
                fmLevel = Math.Abs(fmLevel) * fmPolarity
                
            Case ParameterId.PortamentoEnabled
                portamentoEnabled = (value >= 0.5)
                
            Case ParameterId.Polarity
                polarity = (If(value < 0.5, 1, -1))

            Case Else
                Debug.Fail("Unhandled parameter.")

        End Select
    End Sub

    Public ReadOnly Property ParameterCount() As Integer Implements IProgrammable.ParameterCount
        Get
            Return ParameterId.Polarity + 1
        End Get
    End Property

#End Region

#Region "IBendable"

    Public Property PitchBendModulation() As Single Implements IBendable.PitchBendModulation
        Get
            Return m_pitchBendModulation
        End Get
        Set(ByVal value As Single)
            m_pitchBendModulation = Value
        End Set
    End Property

#End Region

End Class
