

Imports System.Diagnostics
Imports System.IO


Public NotInheritable Class WaveWriter
	Implements IDisposable

    Private Const FormatChunkSize As Integer = 16
    Private Const RiffLengthOffset As Integer = 4
    Private Const DataLengthOffset As Integer = 40

    Private waveStream As Stream
    Private blockAlign As Short
    Private averageBytesPerSecond As Integer
    Private disposed As Boolean = False


    Public Sub New(ByVal channelCount As Short, ByVal samplesPerSecond As Integer, ByVal bitsPerSample As Short)
        Initialize(New MemoryStream(), channelCount, samplesPerSecond, bitsPerSample)
    End Sub

    Public Sub New(ByVal output As Stream, ByVal channelCount As Short, ByVal samplesPerSecond As Integer, ByVal bitsPerSample As Short)
        Initialize(output, channelCount, samplesPerSecond, bitsPerSample)
    End Sub


    Private Sub Initialize(ByVal output As Stream, ByVal channelCount As Short, ByVal samplesPerSecond As Integer, ByVal bitsPerSample As Short)
        If output Is Nothing Then
            Throw New ArgumentNullException("output")
        ElseIf Not output.CanWrite Then
            Throw New ArgumentException("The output stream cannot be written to.")
        ElseIf Not output.CanSeek Then
            Throw New ArgumentException("The stream does not support seeking.")
        ElseIf output.Length <> 0 Then
            Throw New ArgumentException("The stream is not empty.")
        ElseIf channelCount < 1 Then
            Throw New ArgumentOutOfRangeException("channelCount")
        ElseIf samplesPerSecond < 1 Then
            Throw New ArgumentOutOfRangeException("samplesPerSecond")
        ElseIf bitsPerSample < 1 Then
            Throw New ArgumentOutOfRangeException("bitsPerSample")
        End If

        Me.waveStream = output

        CalculateBlockAlign(channelCount, bitsPerSample)
        CalcuateAverageBytesPerSecond(samplesPerSecond)

        WriteRiffChunk(channelCount, samplesPerSecond, bitsPerSample)
    End Sub

    Private Sub CalculateBlockAlign(ByVal channelCount As Short, ByVal bitsPerSample As Short)
        If bitsPerSample Mod 8 = 0 Then
            blockAlign = CShort(((bitsPerSample \ 8) * channelCount))
        Else
            blockAlign = CShort(((bitsPerSample \ 8 + 1) * channelCount))
        End If
    End Sub

    Private Sub CalcuateAverageBytesPerSecond(ByVal samplesPerSecond As Integer)
        averageBytesPerSecond = blockAlign * samplesPerSecond
    End Sub

    Private Sub WriteRiffChunk(ByVal channelCount As Short, ByVal samplesPerSecond As Integer, ByVal bitsPerSample As Short)
        WriteRiffChunkID()
        ' ----------------------------------------------------------- placeholder for wave file length
        waveStream.Write(BitConverter.GetBytes(0), 0, 4)
        WriteWaveChunk(channelCount, samplesPerSecond, bitsPerSample)
    End Sub

    Private Sub WriteRiffChunkID()
        waveStream.WriteByte(Asc("R"c))
        waveStream.WriteByte(Asc("I"c))
        waveStream.WriteByte(Asc("F"c))
        waveStream.WriteByte(Asc("F"c))
    End Sub

    Private Sub WriteWaveChunk(ByVal channelCount As Short, ByVal samplesPerSecond As Integer, ByVal bitsPerSample As Short)
        WriteWaveChunkID()
        WriteFormatChunk(channelCount, samplesPerSecond, bitsPerSample)
        WriteDataChunk()
    End Sub

    Private Sub WriteWaveChunkID()
        waveStream.WriteByte(Asc("W"c))
        waveStream.WriteByte(Asc("A"c))
        waveStream.WriteByte(Asc("V"c))
        waveStream.WriteByte(Asc("E"c))
    End Sub

    Private Sub WriteFormatChunk(ByVal channelCount As Short, ByVal samplesPerSecond As Integer, ByVal bitsPerSample As Short)
        WriteFormatChunkID()
        '
        waveStream.Write(BitConverter.GetBytes(FormatChunkSize), 0, 4)
        waveStream.Write(BitConverter.GetBytes(CShort(1)), 0, 2)
        ' ---------------------------------------------------------------- format type (PCM/Uncompressed)
        waveStream.Write(BitConverter.GetBytes(channelCount), 0, 2)
        waveStream.Write(BitConverter.GetBytes(samplesPerSecond), 0, 4)
        waveStream.Write(BitConverter.GetBytes(averageBytesPerSecond), 0, 4)
        waveStream.Write(BitConverter.GetBytes(blockAlign), 0, 2)
        waveStream.Write(BitConverter.GetBytes(bitsPerSample), 0, 2)
    End Sub

    Private Sub WriteFormatChunkID()
        waveStream.WriteByte(Asc("f"c))
        waveStream.WriteByte(Asc("m"c))
        waveStream.WriteByte(Asc("t"c))
        waveStream.WriteByte(Asc(" "c))
    End Sub

    Private Sub WriteDataChunk()
        WriteDataChunkID()
        ' ------------------------------------------------------------------ placeholder for wave data length
        waveStream.Write(BitConverter.GetBytes(0), 0, 4)
    End Sub

    Private Sub WriteDataChunkID()
        waveStream.WriteByte(Asc("d"c))
        waveStream.WriteByte(Asc("a"c))
        waveStream.WriteByte(Asc("t"c))
        waveStream.WriteByte(Asc("a"c))
    End Sub

    Private Sub PadWaveData()
        If waveStream.Length Mod 2 <> 0 Then
            waveStream.Seek(waveStream.Length, SeekOrigin.Begin)
            ' --------------------------------------------- if data length is not word aligned, write an extra byte.
            Write(0)
        End If
        Debug.Assert(waveStream.Length Mod 2 = 0, "Length not word aligned.")
    End Sub


    Public Sub Write(ByVal sample As Byte)
        If disposed Then
            Throw New ObjectDisposedException(Me.[GetType]().Name)
        End If
        waveStream.WriteByte(sample)
    End Sub

    Public Sub Write(ByVal samples As Byte())
        If disposed Then
            Throw New ObjectDisposedException(Me.[GetType]().Name)
        ElseIf samples Is Nothing Then
            Throw New ArgumentNullException("samples")
        End If
        waveStream.Write(samples, 0, samples.Length)
    End Sub

    Public Sub Close()
        Dispose()
    End Sub


    ' ------------------------------------------------------- gets the underlying stream of the WaveWriter
    Public ReadOnly Property BaseStream() As Stream
        Get
            Return waveStream
        End Get
    End Property



#Region "IDisposable Members"

    Public Sub Dispose() Implements IDisposable.Dispose
        If disposed Then
            Return
        End If

        ' Seek to the RIFF length field.
        waveStream.Seek(RiffLengthOffset, SeekOrigin.Begin)

        ' Write the length of the wave file not counting the RIFF chunk 
        ' ID and the RIFF length field.
        waveStream.Write(BitConverter.GetBytes(waveStream.Length - (RiffLengthOffset + 4)), 0, 4)

        ' Seek to the data length field.
        waveStream.Seek(DataLengthOffset, SeekOrigin.Begin)

        ' Write the length of the data chunk not counting the data chunk 
        ' ID and the data length field.
        waveStream.Write(BitConverter.GetBytes(waveStream.Length - (DataLengthOffset + 4)), 0, 4)

        ' Ensure that the wave file is word aligned - this step should be last before closing.
        PadWaveData()

        waveStream.Close()

        disposed = True

        GC.SuppressFinalize(Me)
    End Sub

#End Region

End Class
